/*
 * Wine debugger GDI Object interrogation routines
 *
 * Copyright 2003 Ulrich Czekalla for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "windef.h"
#include "winbase.h"
#include "wingdi.h"
#include "winuser.h"
#include "tlhelp32.h"
#include "debugger.h"
#include "expr.h"
#include "wine/windef16.h"
#include "gdi.h"

typedef struct
{
    WORD addr;                /* Address of the MOVEABLE block */
    BYTE flags;               /* Flags for this block */
    BYTE lock;                /* Lock count */
} LOCALHANDLEENTRY;

#define HANDLE_MOVEABLE(handle) (((handle) & 3) == 2)
#define FIRST_LARGE_HANDLE 16
#define MAX_LARGE_HANDLES ((GDI_HEAP_SIZE >> 2) - FIRST_LARGE_HANDLE)

typedef struct {
    INT size;
    INT numRects;
    RECT *rects;
    RECT extents;
} WINEREGION;

/* GDI logical region object */
typedef struct
{
    GDIOBJHDR   header;
    WINEREGION  *rgn;
} RGNOBJ;

/* GDI logical brush object */
typedef struct
{
    GDIOBJHDR header;
    LOGBRUSH  logbrush;
} BRUSHOBJ;

static void *DEBUG_GDI_GetObjPtr(HGDIOBJ handle, WORD magic);
static void DEBUG_InfoRegion(GDIOBJHDR *ptr);
static void DEBUG_InfoBitmap(GDIOBJHDR *ptr);
static void DEBUG_InfoBrush(GDIOBJHDR *ptr);

#define GDI_HEAP_SIZE 0xffe0

/***********************************************************************
 *           DEBUG_InfoGDIObj
 *
 * Print information about GDI object
 */
void DEBUG_InfoGDIObj(HGDIOBJ handle)
{
    GDIOBJHDR gdiobj;
    GDIOBJHDR *ptr;
   
    ptr = DEBUG_GDI_GetObjPtr(handle, MAGIC_DONTCARE);

    if (ptr)
    {
        DEBUG_READ_MEM(ptr, &gdiobj, sizeof(GDIOBJHDR));

        switch (GDIMAGIC(gdiobj.wMagic))
	{
            case REGION_MAGIC:
                DEBUG_InfoRegion(ptr);
                break;

	    case BITMAP_MAGIC:
                DEBUG_InfoBitmap(ptr);
                break;

	    case BRUSH_MAGIC:
                DEBUG_InfoBrush(ptr);
                break;

            default:
                DEBUG_Printf("wMagic=%04x\n", GDIMAGIC(gdiobj.wMagic));
	}
    }
    else
        DEBUG_Printf("Invalid GDI handle %p\n", handle);
}


/***********************************************************************
 *           DEBUG_LOCAL_InternalLock
 */
static HLOCAL16 DEBUG_LOCAL_InternalLock( LPSTR heap, HLOCAL16 handle )
{
    if (HANDLE_MOVEABLE(handle))
    {
        LOCALHANDLEENTRY *pEntry;
        LOCALHANDLEENTRY entry;

        pEntry = (LOCALHANDLEENTRY *) (heap + handle);
        DEBUG_READ_MEM(pEntry, &entry, sizeof(LOCALHANDLEENTRY));

        if (entry.flags == (LMEM_DISCARDED >> 8)) return 0;

        handle = entry.addr;
    }

    return handle;
}


/***********************************************************************
 *           DEBUG_LOCAL_Lock
 */
static void *DEBUG_LOCAL_Lock( HANDLE16 ds, HLOCAL16 handle )
{
    void *retptr = NULL;

    if (handle)
    {
        DBG_VALUE rtn;

        if (DEBUG_GetSymbolValue("wine_ldt_copy", -1, &rtn, FALSE) == gsv_found)
        {
            char *ptr;
            BYTE *ldt;
            DWORD segptr;

            ldt = (void*)DEBUG_ToLinear(&rtn.addr);
            segptr = MAKESEGPTR( ds, 0 );

            DEBUG_READ_MEM(ldt + ((SELECTOROF(segptr) >> __AHSHIFT) * sizeof(void*)), 
                &ptr, sizeof(char*));
            ptr += OFFSETOF(segptr);

            retptr = ptr + DEBUG_LOCAL_InternalLock( ptr, handle );
        }
    }

    return retptr;
}


/***********************************************************************
 *           DEBUG_GDI_GetObjPtr
 */
static void *DEBUG_GDI_GetObjPtr( HGDIOBJ handle, WORD magic )
{
    GDIOBJHDR *ptr = NULL;

    if ((UINT_PTR)handle & 2)  /* GDI heap handle */
    {
        DBG_VALUE rtn;
        HLOCAL16 h = LOWORD(handle);

        if (DEBUG_GetSymbolValue("GDI_HeapSel", -1, &rtn, FALSE) == gsv_found)
        {
            DWORD psel;
            WORD gdi_heapsel;

            psel = DEBUG_ToLinear(&rtn.addr);
            DEBUG_READ_MEM((LPCVOID)psel, &gdi_heapsel, sizeof(gdi_heapsel));

            ptr = (GDIOBJHDR *)DEBUG_LOCAL_Lock(gdi_heapsel, h);
            if (ptr)
            {
                GDIOBJHDR gdiobj;

                DEBUG_READ_MEM(ptr, &gdiobj, sizeof(GDIOBJHDR));

                if (((magic != MAGIC_DONTCARE) && (GDIMAGIC(gdiobj.wMagic) != magic)) ||
                   (GDIMAGIC(gdiobj.wMagic) < FIRST_MAGIC) ||
                   (GDIMAGIC(gdiobj.wMagic) > LAST_MAGIC))
                {
                    ptr = NULL;
                }
            }
        }
    }
    else  /* large heap handle */
    {
        int i = ((UINT_PTR)handle >> 2) - FIRST_LARGE_HANDLE;
        if (i >= 0 && i < MAX_LARGE_HANDLES)
        {
            DBG_VALUE rtn;

            if (DEBUG_GetSymbolValue("large_handles", -1, &rtn, FALSE) == gsv_found)
            {
                char* large_handles = (void*)DEBUG_ToLinear(&rtn.addr);

                DEBUG_READ_MEM(large_handles + (i * sizeof(GDIOBJHDR*)), 
                    &ptr, sizeof(GDIOBJHDR*));

                if (ptr)
                {
                    GDIOBJHDR gdiobj;

		    DEBUG_READ_MEM(ptr, &gdiobj, sizeof(GDIOBJHDR));

                    if ((magic != MAGIC_DONTCARE) && (GDIMAGIC(gdiobj.wMagic) != magic))
                        ptr = NULL;
                }
	    }
        }
    }

    return ptr;
}


/***********************************************************************
 *           DEBUG_GetRegionData
 */
static DWORD DEBUG_GetRegionData(GDIOBJHDR *obj, DWORD count, LPRGNDATA rgndata)
{
    DWORD size;
    RGNOBJ rgnobj;
    WINEREGION winergn;

    if(!obj)
        return 0;

    DEBUG_READ_MEM(obj, &rgnobj, sizeof(RGNOBJ));

    if (!rgnobj.rgn)
        return 0;

    DEBUG_READ_MEM(rgnobj.rgn, &winergn, sizeof(WINEREGION));

    size = winergn.numRects * sizeof(RECT);

    if(count < (size + sizeof(RGNDATAHEADER)) || rgndata == NULL)
    {
	if (rgndata) /* buffer is too small, signal it by return 0 */
	    return 0;
	else		/* user requested buffer size with rgndata NULL */
	    return size + sizeof(RGNDATAHEADER);
    }

    rgndata->rdh.dwSize = sizeof(RGNDATAHEADER);
    rgndata->rdh.iType = RDH_RECTANGLES;
    rgndata->rdh.nCount = winergn.numRects;
    rgndata->rdh.nRgnSize = size;
    rgndata->rdh.rcBound.left = winergn.extents.left;
    rgndata->rdh.rcBound.top = winergn.extents.top;
    rgndata->rdh.rcBound.right = winergn.extents.right;
    rgndata->rdh.rcBound.bottom = winergn.extents.bottom;

    DEBUG_READ_MEM(winergn.rects, rgndata->Buffer, size);

    return size + sizeof(RGNDATAHEADER);
}


/***********************************************************************
 *           DEBUG_InfoRegion
 */
static void DEBUG_InfoRegion(GDIOBJHDR *ptr)
{
    INT i;
    DWORD size;
    LPRECT lprect;
    LPRGNDATA rgndata;

    size = DEBUG_GetRegionData(ptr, 0, NULL);

    if (size)
    {
        rgndata = (LPRGNDATA) HeapAlloc(GetProcessHeap(), 0, size);
	size = DEBUG_GetRegionData(ptr, size, rgndata);

        if (size)
        {
            DEBUG_Printf( "Type=0x%08lx  count=%ld\n"
                          "rcBound=%ld,%ld-%ld,%ld\n",
                          rgndata->rdh.iType,
                          rgndata->rdh.nCount,
                          rgndata->rdh.rcBound.left, rgndata->rdh.rcBound.top,
                          rgndata->rdh.rcBound.right, rgndata->rdh.rcBound.bottom);

            lprect = (LPRECT) rgndata->Buffer;
	    for (i = 0; i < rgndata->rdh.nCount; i++)
                DEBUG_Printf( "Rect%d=%ld,%ld-%ld,%ld\n",
                              i,
                              lprect[i].left, lprect[i].top,
                              lprect[i].right, lprect[i].bottom);
        }

	HeapFree(GetProcessHeap(), 0, rgndata);
    }
}


/***********************************************************************
 *           DEBUG_InfoBitmap
 */
static void DEBUG_InfoBitmap(GDIOBJHDR *ptr)
{
    BITMAPOBJ bmp;

    if (DEBUG_READ_MEM(ptr, &bmp, sizeof(bmp)))
    {
        DEBUG_Printf( "Type=0x%08x\n"
                      "width=%d height=%d widthBytes=%d planes=%d bitsPixel=%d\n",
                      bmp.bitmap.bmType, bmp.bitmap.bmWidth, bmp.bitmap.bmHeight,
                      bmp.bitmap.bmWidthBytes, bmp.bitmap.bmPlanes, bmp.bitmap.bmBitsPixel);
    }
}


/***********************************************************************
 *           DEBUG_InfoBrush
 */
static void DEBUG_InfoBrush(GDIOBJHDR *ptr)
{
    BRUSHOBJ brush;

    if (DEBUG_READ_MEM(ptr, &brush, sizeof(brush)))
    {
        DEBUG_Printf( "lbStyle=%d lbColor=0x%08lx lbHatch=%d\n",
            brush.logbrush.lbStyle, brush.logbrush.lbColor, brush.logbrush.lbHatch);
    }
}
